#!/usr/bin/perl
# 
# plucacheall_pgsql.pl
#
# author Al Kingsley <al@bananahead.com>
# license	GPL version 2 or later.
# copyright Al Kingsley 2004
#
#	Create or Update the PLU cache table
#

use DBI;
use Getopt::Long;
use POSIX qw(strftime);
use Time::Local;
use Cwd;
#use MIME::Lite;
# COMMENT out the use Net::SMTP line you're not using smtp (M$ option)
#use Net::SMTP;

# some options
$email_notification = 'no';
$mailprog           = '/usr/sbin/sendmail -t -oi';
#$mailprog          = '/usr/lib/sendmail -t -oi';
#$mailprog          = "/var/qmail/bin/qmail-inject -h";
$smtp_host          = 'your.mailserver.domain.or.ip';
$send_method        = 'sendmail'; # or 'smtp'
$admin_email_to     = "you\@yourdomain.com";
$admin_email_from   = "bhposadmin\@yourdomain.com";
$subject            = "PLU cache generatation";
$body_text          = '';

$VER = "1.2.2";

$opt_db = "bananas";
$opt_max_size = 65536;
$opt_db_engine = "mysql";
$opt_host = "localhost";
$opt_user = "root";
$opt_password = "bollocks";
$opt_avg_period = 30;
$opt_help = $opt_version = 0;

GetOptions("help","version","user=s","password=s",
	   "db_engine=s","db=s","host=s","max_size=s","avg_period=s") || usage();

usage($VER) if ($opt_help || $opt_version);

$trans_processed = $items_processed = $num_in_cache = 0;

$dbh = DBI->connect("DBI:$opt_db_engine:dbname=$opt_db",$opt_user,
		    $opt_password,{ PrintError => 0}) || die $DBI::errstr;

if(generate_synch_entry($dbh) < 0){
   if($email_notification eq 'yes' ){
      email_notice();
   }
   exit(0);
}
generate_plucache($dbh);
close_synch_entry($dbh);
$body_text.="Success: cached $num_in_cache products\n";
if($email_notification eq 'yes' ){
   email_notice();
}
print "total cached:\t\t\t$num_in_cache\n"; 
print "\n";

$dbh->disconnect;
exit(0);

sub generate_plucache
{
	my($dbh)=@_;
	my($sth,$tth,$query);
	%cache_entry=();
	
	$opt_avg_period+=1;

	print "Generating PLU cache...\n";

	new_plucache_table($dbh);

	$query = "select id, plu from product order by plu";
	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	while(my @r = $sth->fetchrow_array){
		$cache_entry{'id'}='';	 
      $cache_entry{'productid'}=$r[0];
      $cache_entry{'nameid'}=1;
	   $cache_entry{'plu'}=$r[1];
		$cache_entry{'disptext'}='';
		$cache_entry{'recutext'}='';
		$cache_entry{'unitprice'}='0.00';
		$cache_entry{'discount'}='0.00';
      $cache_entry{'deposittype'}='0';
		$cache_entry{'deposit'}='0.00';
		$cache_entry{'tax'}='0.00';
		$cache_entry{'itemflags'}='0';
		
		$cache_entry=get_product($dbh, $r[0], \%cache_entry);
		
		#$num_in_cache++;
	}
}

sub average_table
{
	my($dbh)=@_;
	my($sth,$tth,$query);

	$query  = "SELECT * FROM temp_plu";
	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	while(my @r = $sth->fetchrow_array){
		$r[4]=($r[3]/30);
		print "item: $r[0]- $r[4]\n";
		$tth = $dbh->prepare("UPDATE temp_plu SET avg='$r[4]' WHERE id = '$r[0]'") || die $dbh->errstr;
		$tth->execute || die $dbh->errstr;
	}
}

sub create_temp_table
{
	my($dbh)=@_;
	my($sth,$create);
  
	$sth = $dbh->prepare("select count(*) from temp_plu") or die $dbh->errstr;
	if($sth->execute){
		$create = "DELETE FROM temp_plu";
		$sth = $dbh->prepare($create) or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
	}
	else{
		$create = "CREATE TABLE temp_plu(";
		$create .= "id int(11) NOT NULL auto_increment, ";
		$create .= "pid int NOT NULL, ";
		$create .= "qty int(11) NOT NULL, ";
		$create .= "peak int(11) NOT NULL, ";
		$create .= "avg float NOT NULL, ";
		$create .= "PRIMARY KEY (id)) ";

		$sth = $dbh->prepare($create) or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
	}
}

sub query_trans
{
	my($dbh)=@_;
	my($sth,$query);

	$query  = "select id, transactionid, entereddate from transaction where ";
	$query .= "((to_days(now()) - 30) > (to_days(entereddate)/365)) ";
	$query .= "order by entereddate desc";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	
	my $count=$sth->rows;
	
	print "Processing $count Transactions...\n";

	while(my @r = $sth->fetchrow_array){
		process_transitems($dbh, $r[0]);
		$trans_processed++;
	}
}

sub process_transitems
{
	my($dbh, $tid)=@_;
	my($sth,$query);

	$query = "select * from transactionitems where transactionid = '$tid' ORDER BY quantity";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	
	my $count=$sth->rows;
	
	print "Processing Transaction $tid : $count items\n";
	
	while(my @row = $sth->fetchrow_array){
		process_product($dbh, $row[3], $row[4]);
	}
	
	print "\t\tDone.\n";
}

sub process_product
{
	my($dbh,$pid,$qty)=@_;
	my($sth,$query, $row);

	$query = "SELECT * FROM temp_plu WHERE pid = '$pid'";
	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	my @r = $sth->fetchrow_array;
   if($r[0]){	# got a row
		$r[2] = $r[2] + $qty;
		if($qty > $r[3]){
			$r[3]=$qty;
		}
		print "repeat product: $pid, qty=$r[2]\n";
		
		$sth = $dbh->prepare("UPDATE temp_plu SET qty=$r[2], peak=$r[3] WHERE ID = $r[0]") || die $dbh->errstr;
		$sth->execute || die $dbh->errstr;
	}
	else{
		print "product: $pid, qty=$qty\n";
		$query = "INSERT INTO temp_plu VALUES('0', '$pid', '$qty', '$qty', '0.0')"; 
		$sth = $dbh->prepare($query) || die $dbh->errstr;
		$sth->execute || die $dbh->errstr;
		$items_processed++;
	}
}

sub get_product
{
	my($dbh,$pid,$cache_entry)=@_;
	my($sth,$query);

	$query  = "SELECT id, plu, name, reciepttext, sellquantityid, promotionid, flags, description, depositid, discountid from product where id = $pid";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
   
	while(my @row = $sth->fetchrow_array){
      $cache_entry->{'productid'} = $row[0];
		$cache_entry->{'plu'} = $row[1];
		
		if($r[1]){
		   $cache_entry->{'disptext'} = $row[2];
		}
		else{
		   $cache_entry->{'disptext'} = $row[7];
		}
      if(length($cache_entry->{'disptext'}) > 64){
         $cache_entry->{'recutext'}=substr($cache_entry->{'recutext'}, 0, 64);
      }
      
      $cache_entry->{'recutext'} = $row[3];
      if(length($cache_entry->{'recutext'}) > 32){
         $cache_entry->{'recutext'}=substr($cache_entry->{'recutext'}, 0, 32);
      }
		if($r[5]){
			$cache_entry->{'itemflags'} = $row[6];
		}
		else{
			$cache_entry->{'itemflags'} = "0";
		}
		if($r[4]){
			#$cache_entry=get_promo($dbh, $row[4], $cache_entry);
         $cache_entry=get_discount($dbh, $row[9], $cache_entry);
		}
		else{
			$cache_entry->{'discount'} = "0.00";
		}
      if($row[8]){
         $cache_entry=get_deposit($dbh, $row[8], $cache_entry);
		}
		else{
         $cache_entry->{'deposittype'} = "0";
			$cache_entry->{'deposit'} = "0.00";
		}
      $cache_entry=get_nameprice($dbh, $row[4], $cache_entry);
		return $cache_entry;
	}
}

sub get_deposit
{
   my($dbh,$did,$cache_entry)=@_;
	my($sth,$query);

	$query = "SELECT percentage, type FROM deposit WHERE id = '$did'";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	
	if(my @row = $sth->fetchrow_array){
		$cache_entry->{'deposit'} = $row[0];
      
      if($row[1] eq 'rental'){
      }
      if($row[1] eq 'layby'){
      }
      if($row[1] eq 'returnable'){
      }
      if($row[1] eq 'redemption'){
         $cache_entry->{'deposittype'} = '2';
      }
	}
	return $cache_entry;
}

sub get_price
{
	my($dbh,$pid,$cache_entry)=@_;
	my($sth,$query);

	$query = "select price, taxid from prodpricequantity where pqid = '$pid' AND nameid='1'";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	
	if(my @row = $sth->fetchrow_array){
		$cache_entry->{'unitprice'} = $row[0];
		$cache_entry=get_tax($dbh, $row[1], $cache_entry);
	}
	return $cache_entry;
}

sub get_nameprice
{
	my($dbh,$pid,$cache_entry)=@_;
	my($sth,$query);

	$query = "select price, taxid, nameid from prodpricequantity where pqid = '$pid'";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	
	while(my @row = $sth->fetchrow_array){
      $cache_entry->{'unitprice'} = $row[0];
      $cache_entry->{'nameid'} = $row[2];
		$cache_entry=get_tax($dbh, $row[1], $cache_entry);
      
      $query  = "INSERT INTO plu_cache VALUES('', ";
      $query .= $dbh->quote($cache_entry->{'productid'}) . ",";
      $query .= $dbh->quote($cache_entry->{'nameid'}) . ",";
		$query .= $dbh->quote($cache_entry->{'plu'}) . ",";
		$query .= $dbh->quote($cache_entry->{'disptext'}) . ",";
		$query .= $dbh->quote($cache_entry->{'recutext'}) . ",";
		$query .= $dbh->quote($cache_entry->{'unitprice'}) . ",";
		$query .= $dbh->quote($cache_entry->{'discount'}) . ",";
      $query .= $dbh->quote($cache_entry->{'deposittype'}) . ",";
		$query .= $dbh->quote($cache_entry->{'deposit'}) . ",";
		$query .= $dbh->quote($cache_entry->{'tax'}) . ",";
		$query .= $dbh->quote($cache_entry->{'itemflags'}) . ")";
      
		$tth = $dbh->prepare($query) || die $dbh->errstr;
		$tth->execute || die $dbh->errstr;
		
		$num_in_cache++;
	}
	return $cache_entry;
}

sub get_discount
{
	my($dbh,$did,$cache_entry)=@_;
	my $query  = "select margin from discount where discountid = '$did'";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if(my @row = $sth->fetchrow_array){
		$cache_entry->{'Discount'} = $row[0];
	}
	else{
		$cache_entry->{'Discount'} = "0.00";
	}
	return $cache_entry;
}

sub get_promo
{
	my($dbh,$pid,$cache_entry)=@_;
	my $query  = "select margin from promotion where promotionid = '$pid' AND TO_DAYS(enddate) < TO_DAYS(NOW())";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if(my @row = $sth->fetchrow_array){
		$cache_entry->{'discount'} = $row[0];
	}
	else{
		$cache_entry->{'discount'} = "0.00";
	}
	return $cache_entry;
}

sub get_tax
{
	my($dbh,$tid,$cache_entry)=@_;	my($sth,$query);

	$query  = "SELECT rate FROM taxrate WHERE id = '$tid'";
	
	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if(my @row = $sth->fetchrow_array){
		$cache_entry->{'Tax'} = $row[0];
	}
	else{
	   $query  = "SELECT rate FROM taxrate WHERE id = '1'";
	
   	$sth = $dbh->prepare($query) or die $dbh->errstr;
   	$sth->execute() or die $dbh->errstr;
   	if(my @row = $sth->fetchrow_array){
   		$cache_entry->{'Tax'} = $row[0];
   	}
	}
	return $cache_entry;
}

sub new_plucache_table
{
	my($dbh)=@_;
	my($sth,$query);
	
	$sth = $dbh->prepare("select count(*) from plu_cache") or die $dbh->errstr;
	if($sth->execute){
		$create = "DELETE FROM plu_cache";
		$sth = $dbh->prepare($create) or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
	}
   else{
      $query = "CREATE TABLE plu_cache (";
	   $query .= "id int(11) not null serial,";
      $query .= "productid int(11) default '0' not null,";
      $query .= "nameid int(11) default '1' not null,";
	   $query .= "plu varchar(16) not null,";
	   $query .= "disptext varchar(64) not null,";
	   $query .= "recutext varchar(32) not null,";
	   $query .= "unitprice float not null default '0.0',";
 	   $query .= "discount float not null default '0.0',";
      $query .= "deposittype int not null default '0',";
 	   $query .= "deposit float not null default '0.0',";
 	   $query .= "tax float not null default '0.0',";
 	   $query .= "itemflags int(10) unsigned default '0' not null,";
 	   $query .= "PRIMARY KEY (id))";

		$sth = $dbh->prepare($query) or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
	}
}

# put the synch entry in on start in case of interrupts
sub generate_synch_entry
{
	my($dbh)=@_;
	my($sth);

	$sth = $dbh->prepare("SELECT * FROM synch WHERE implemented='N' AND typeid='1'") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows){
		print "The synch table has outstanding entries, cannot perform operation\n";
		$body_text.="The synch table has outstanding entries, cannot perform operation\n";
		return -1;
	}
	$sth = $dbh->prepare("INSERT INTO synch VALUES('','1',NOW(),NOW(),'epoch','N')") || die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
   return 0;
}

sub get_date
{
	my($dbh,$ts)=@_;
	my($sth,$query);

	$sth = $dbh->prepare("SELECT implementationat from synch where implemented='Y' and typeid='1' ORDER BY  implementationat DESC LIMIT 1") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows==0){
		print "The synch table has no last recstock entry - quitting\n";
      $body_text.="The synch table has outstanding entries, cannot perform operation\n";
		return -1;
	}
	my @r = $sth->fetchrow_array;
	return $r[0];
}

# close synch entry
sub close_synch_entry
{
	my($dbh)=@_;
	my($sth, $tth);

	$sth = $dbh->prepare("SELECT ID FROM synch WHERE implemented='N' AND typeid='1'") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows){
      my @r = $sth->fetchrow_array;
		$tth = $dbh->prepare("UPDATE synch SET implemented='Y', implementationat=NOW() WHERE id= $r[0]") || die $dbh->errstr;
	   $tth->execute || die $dbh->errstr;
	}
}

# new option
sub generate_new_synch_entry
{
	my($dbh)=@_;
	my($sth);

	$sth = $dbh->prepare("INSERT INTO synch VALUES('','1',NOW(),NOW(),NOW(),'Y')") || die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
}

sub email_notice
{
   if($send_method eq 'smtp'){
      # use this for windows machines that don't have sendmail
      MIME::Lite->send("$send_method", "$mailprog_or_smtp_host", Timeout=>60);
      $msg = new MIME::Lite
             From     =>"$admin_email_from",
             To       =>"$admin_email_to",
             Subject  =>"$subject",
             Type     =>"TEXT",
             Encoding =>"7bit",
             Data     =>"$body_text"; 
  
      $msg->send || die print qq~Error!\n\nError in Mailing Program!~;
   }
   else{
      &mail_to($admin_email_to, $admin_email_from, $subject, $body_text, $admin_email_from);
   }
}

sub mail_to
{
   # &mail_to($email_to, $email_from, $subject, $mail_body, $reply_to);
   my($email_to, $email_from, $subject, $mail_body, $reply_to) = @_;

   print "Sending email notification\n";
   
   if($reply_to !~ /\@/ ){$reply_to = $email_from;}

   open(MAIL, "|$mailprog") || die print qq~Error!\n\nCan't open $mailprog!~;

   print MAIL "To: $email_to\n";
   print MAIL "From: $email_from\n";
   print MAIL "Subject: $subject\n";
   print MAIL "Reply-To: $reply_to\n";
   print MAIL "\n";
   print MAIL "$mail_body";
   print MAIL "\n";
   close (MAIL);
}

sub usage
{  
  my($VER)=@_;
  
  $0 =~ s/.\/(.+)/$1/;
  if ($opt_version)
  {
    print "$0 version $VER\n";
  } 
  else
  {
    print <<EOF;
$0 version $VER

Usage: $0 [options] 

Description: Create a new PLU cache table

Options:
--help             show this help and exit
--version          shows the version of the program
--db_engine=...    database server (default: $opt_db_engine)
--db=...           database to be used (default: $opt_db)
--host=...         hostname to be used (default: $opt_host)
--password=...     user password for the db server
--user=...         username for the db server
--avg_period=...   number of days to average over
--max_size=#		 max number of PLUs to store in cache (default $opt_max_size)
EOF
  }
  exit(0);
}
